# License:
#   Copyright (C) 2011 Thomas GUILLET, Damien CHAPON, Marc LABADENS. All Rights Reserved.
#
#   This file is part of PyMSES.
#
#   PyMSES is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   PyMSES is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with PyMSES.  If not, see <http://www.gnu.org/licenses/>.
""" ray_trace.pyx -- Ray-tracer routine
"""
# cython: profile=True

cimport numpy
cimport cython
import numpy
from pymses.sources.ramses.hilbert import indices_to_positions
from pymses.sources.ramses.tree_utils import tree_search

# Define integer and double precision float types
INT_t = numpy.int
ctypedef numpy.int_t cINT_t
DBL_t = numpy.float64
ctypedef numpy.float64_t cDBL_t


@cython.boundscheck(False)
def ray_trace_amr(dset, numpy.ndarray[cDBL_t, ndim=2] ray_origins, \
		numpy.ndarray[cDBL_t, ndim=2] ray_vects, \
		cDBL_t depth_max, op, info):
	""" Ray tracer

	ray_origins --

	ray_vects --

	depth_max --

	op --

	info --
	"""
	# integer variables
	cdef cINT_t nrays= ray_origins.shape[0]
	cdef cINT_t ndim_rays = ray_origins.shape[1]
	cdef cINT_t ndim = dset.amr_header["ndim"]
	assert (ndim_rays == ndim)
	cdef cINT_t max_read_level = dset.amr_struct["readlmax"]
	cdef cINT_t nfunc = op.nscal_func()
	cdef cINT_t ngrids = dset.amr_struct["ngrids"]
	cdef cINT_t twotondim = (1 << ndim)
	cdef cINT_t psize = (twotondim-1) * max_read_level
	cdef cINT_t iray, idim, iblock, nblocks, level, ilevel, ifunc
	cdef cINT_t igrid_root, icell_root, igrid, icell, igrid_son, dhead
	cdef cINT_t ind_son, head, i, intersect, iface, sgn_face, idim2
	cdef cINT_t ind_son1, ind_son2, ind_entry, ind_exit, add#, ncd, ncdnul
	cdef int max_op = op.is_max_alos()
	cdef int use_dx = op.use_cell_dx()

	# float variables
	cdef cDBL_t alpha, alpha_save, dx, cube_depth, r2, l, d, v
	cdef cDBL_t Rsphere, Rsphere2
	cdef cDBL_t sqrt3 = numpy.sqrt(3.0)
	cdef cDBL_t dm2 = depth_max/2.0
	
	# Big AMR arrays
	cdef numpy.ndarray[int, ndim=1] igrids = numpy.empty(ngrids, 'i')
	cdef numpy.ndarray[char, ndim=1] icells = numpy.empty(ngrids, 'i1')
	cdef numpy.ndarray[int, ndim=2] sons = dset.amr_struct["son_indices"]
	cdef numpy.ndarray[int, ndim=1] g_levels =  dset.get_grid_levels()
	cdef numpy.ndarray[cDBL_t, ndim=3] cell_centers = dset.get_cell_centers()

	# Pile arrays	
	cdef numpy.ndarray[cINT_t, ndim=1] igrid_pile = numpy.empty(psize, INT_t)
	cdef numpy.ndarray[cINT_t, ndim=1] icell_pile = numpy.empty(psize, INT_t)
	cdef numpy.ndarray[cINT_t, ndim=1] level_pile = numpy.empty(psize, INT_t)

	# Integrand variables
	cdef numpy.ndarray[cDBL_t, ndim=3] scal_data = numpy.empty((ngrids, twotondim, nfunc), DBL_t)
	# Operand data
	ifunc = 0
	for (key, scal_func) in op.iter_scalar_func():
		scal_data[:,:,ifunc] = scal_func(dset)
		ifunc += 1
	
	# (ndim,) vector arrays
	cdef numpy.ndarray[cDBL_t, ndim=1] block_center = numpy.empty(ndim, DBL_t)
	cdef numpy.ndarray[cDBL_t, ndim=1] fc = numpy.empty(ndim, DBL_t)
	cdef numpy.ndarray[cDBL_t, ndim=1] I = numpy.empty(ndim, DBL_t)
	cdef numpy.ndarray[cINT_t, ndim=1] dim_constr = numpy.empty(ndim, INT_t)
	cdef numpy.ndarray[cINT_t, ndim=1] dim_side = numpy.empty(ndim, INT_t)
	
	# Output maps
	cdef numpy.ndarray[cDBL_t, ndim=2] maps = numpy.zeros((nrays, nfunc), DBL_t)


	cdef numpy.ndarray[cINT_t, ndim=1] order


	ind_list, order_list = info["dom_decomp"].minimal_domain(dset.icpu)
	order = order_list.astype(INT_t)
	pos_blocks = (indices_to_positions(ind_list, order_list)).astype(DBL_t)
	nblocks = pos_blocks.shape[0]

	search_dict = tree_search(dset.amr_struct, pos_blocks, order)
	igrids = search_dict["grid_indices"]
	icells = search_dict["cell_indices"]

#	ncd = 0
#	ncdnul = 0
	# iteration over the minimal grid description of the domain
	for iblock in range(nblocks):
		igrid_root = igrids[iblock]
		icell_root = icells[iblock]
		level = g_levels[igrid_root]
		dx = 1.0/(1<<level)
		Rsphere = sqrt3 * dx / 2.0
		Rsphere2 = Rsphere*Rsphere
		# Block center (ndim,) coordinates
		for idim in range(ndim):
			block_center[idim] = cell_centers[igrid_root, icell_root, idim]
		#################
		# Ray iteration #
		#################
		for iray in range(nrays):
			# Coordinate parameter of the point of the ray nearest to the block center
			alpha = 0.0
			for idim in range(ndim):
				alpha += (block_center[idim]-ray_origins[iray,idim])*ray_vects[iray,idim]
			# square distances of the ray from the block center
			r2 = 0.0
			for idim in range(ndim):
				v = ray_origins[iray,idim] + alpha * ray_vects[iray, idim] - block_center[idim]
				r2 += v*v
			# Is this ray able to intersect the block ?
			if (r2 > Rsphere2): # no => continue
				continue
			# distance from the ray extreme points
			d = alpha-dm2
			if d<0.: d=-d
			d = d-dm2
			if d>0.:
				r2 += d*d
				# Is this ray still able to intersect the block ?
				if (r2 > Rsphere2):
					continue
			# Pile initialisation with current block
			head=0
			igrid_pile[head] = igrid_root
			icell_pile[head] = icell_root
			level_pile[head] = level
			while head !=-1 :
				igrid = igrid_pile[head]
				icell = icell_pile[head]
				ilevel = level_pile[head]
				head -= 1
				# Current cell size
				dx = 1.0/(1<<ilevel)
				# Cube optical depth computation
				cube_depth = 0.0
				intersect = 0
				for idim2 in range(ndim):
					if intersect==2: break
					for sgn_face in range(-1,2,2):
						if(ray_vects[iray, idim2] != 0.0): # ray is not parallel to the
						# faces along dimension idim. There might be an intersection.
							# Coordinate parameter of the intersection of the ray with the face plane
							alpha = ((cell_centers[igrid, icell, idim2] + sgn_face * dx / 2.0) \
									- ray_origins[iray, idim2])/ray_vects[iray, idim2]
							# Is this intersection within the face ?
							v=0.0
							ind_son = (twotondim-1) - (((-sgn_face+1)>>1)<<idim2)
							for idim in range(ndim):
								if idim != idim2:
									l = (ray_origins[iray,idim] + alpha * ray_vects[iray,idim] - cell_centers[igrid, icell, idim])/dx
									if l<0.0:
										l=-l
										ind_son -= (1<<idim)
									if l>v: v=l
							if v < 0.5:
								# The intersection is in the face of the cube
								intersect +=1
								if intersect == 1: # First intersection found
									if alpha < 0.0:
										alpha_save = 0.0
									elif alpha > depth_max:
										alpha_save = depth_max
									else:
										alpha_save = alpha
									ind_son1=ind_son
								else: # Second intersection found : done
									if alpha < 0.0:
										alpha = 0.0
									elif alpha > depth_max:
										alpha = depth_max
									cube_depth = alpha-alpha_save
									# Make sure ind_son1 = entry point son index
									#           ind_son2 = exit point son index
									if cube_depth < 0.0 : # ind_son is the entry point son index
										cube_depth=-cube_depth
										ind_son2 = ind_son1
										ind_son1 = ind_son
									else:# ind_son is the exit point son index
										ind_son2=ind_son
									break

				# Is the current cell actually intersected by the ray ?
				if cube_depth > 0.0:
#					ncd += 1
					# The ray cuts the block
					igrid_son = sons[igrid, icell]
					if (igrid_son < 0) or (ilevel >= max_read_level):
					# We're done searching: the subcell is the one we want
						for ifunc in range(nfunc):
							if use_dx:
								v = ilevel * 1.0
							else:
								v = scal_data[igrid, icell, ifunc]
							if max_op:
								if v>maps[iray, ifunc]:
									maps[iray, ifunc] = v
							else:
								maps[iray, ifunc] += v * cube_depth

					else: # Go down the AMR tree
						#########################################################################################
						#                      Add the NECESSARY son cells to the pile                          # 
						#########################################################################################
						if ind_son1 != ind_son2:
							for idim in range(ndim):
								ind_entry = (ind_son1>>idim) & 1
								ind_exit = (ind_son2>>idim) & 1
								if (ind_entry==ind_exit):
									dim_constr[idim] = 1
									dim_side[idim] = ind_entry
								else:
									dim_constr[idim] = 0

							# Add the POSSIBLY intersecting son cells different from the entry/exit point son cells
							for ind_son in range(twotondim): # All he son cells
								add = 1
								# Different from entry/exit point son cells
								if (ind_son == ind_son1) or (ind_son== ind_son2):
									add = 0
								else:
									for idim in range(ndim):
										if dim_constr[idim] and (dim_side[idim] != (ind_son>>idim) & 1):
											add = 0
											break

								if add:
									head += 1
									level_pile[head] = ilevel+1
									igrid_pile[head] = igrid_son
									icell_pile[head] = ind_son

							# If different from the son cell of the entry point, add the exit point son cell
							head+=1
							level_pile[head] = ilevel+1
							igrid_pile[head] = igrid_son
							icell_pile[head] = ind_son2
						
						# Add the son cell of the entry point of the ray in the father grid
						head+=1
						level_pile[head] = ilevel+1
						igrid_pile[head] = igrid_son
						icell_pile[head] = ind_son1
#				else:
#					ncdnul += 1
						
						#########################################################################################

#	print ncd, ncdnul
	return maps
