# License:
#   Copyright (C) 2011 Thomas GUILLET, Damien CHAPON, Marc LABADENS. All Rights Reserved.
#
#   This file is part of PyMSES.
#
#   PyMSES is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   PyMSES is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with PyMSES.  If not, see <http://www.gnu.org/licenses/>.
""" tree_utils.py -- low-level tree utility functions
"""

import numpy
cimport numpy
cimport cython


@cython.boundscheck(False)
def tree_search(amr_struct, numpy.ndarray[double, ndim=2]points,
		max_search_level=None):
	""" Searches deepest cells containing points in a RAMSES tree structure.

	This function searches the deepest cells containing the points `points`,
	and returns the corresponding grid and cell indices.

	Parameters:

		amr_struct -- a dictionary tuple containing the AMR structure, such as
			the output of read_ramses_amr_file()[1]

		points -- a (npoints, ndim) numpy array holding the coordinates of the
		    points to search for, which should be in [0, 1[^ndim

		max_search_level -- the deepest level to explore (int or a (npoints,) numpy array of int

	Output:

		a dictionary containing info about the matching cells.
	"""

	points = numpy.asarray(points)

	cdef int npoints = points.shape[0]
	cdef char ndim = points.shape[1]
	cdef char max_read_level = amr_struct["readlmax"]

	cdef char c_max_search_level
	cdef numpy.ndarray[char, ndim=1] c_max_search_level_array
	if max_search_level is None:
		c_max_search_level_array = max_read_level * \
				numpy.ones(npoints, 'i1')
	else:
		if isinstance(max_search_level, numpy.ndarray):
			assert max_search_level.size == npoints
			c_max_search_level_array = max_search_level.astype('i1')
		if isinstance(max_search_level, list):
			assert len(max_search_level)==npoints
			c_max_search_level_array = numpy.asarray(max_search_level, 'i1')


	# Big AMR arrays
	cdef numpy.ndarray[double, ndim=2] xgrids = amr_struct["grid_centers"]
	cdef numpy.ndarray[int, ndim=2] sons = amr_struct["son_indices"]

	# Output arrays
	cdef numpy.ndarray[int, ndim=1] point_igrids = numpy.empty(npoints, 'i')
	cdef numpy.ndarray[char, ndim=1] point_icells = numpy.empty(npoints, 'i1')
	cdef numpy.ndarray[char, ndim=1] point_ilevels = numpy.empty(npoints, 'i1')

	cdef int ipoint, igrid, igridstart
	cdef char ibit, idim, ind, ilevel

	# This was in io_ramses.f90, but does not work
	#igridstart = amr_struct["coarse_son_indices"][0]
	igridstart = 0

	# Loop over points
	for ipoint in range(npoints):
		c_max_search_level = min(c_max_search_level_array[ipoint], max_read_level)

		# Start search with the topmost grid

		# Was in ioramses.f90: igrid = region_sons[0]
		igrid = igridstart

		ilevel = 1
		# Search down levels
		for ilevel in range(1, c_max_search_level+1):
			# Find the subcell
			ind = 0
			for idim in range(ndim):
				ibit = (points[ipoint, idim] > xgrids[igrid, idim])*1
				ind += (ibit << idim)

			# Is the subcell a leaf cell?
			if sons[igrid, ind] < 0 or ilevel == c_max_search_level:
				# We're done searching: the subcell is the one we want
				break
			else:
				# Go down the tree
				igrid = sons[igrid, ind]

		point_igrids[ipoint] = igrid
		point_icells[ipoint] = ind
		point_ilevels[ipoint] = ilevel

	return { "grid_indices" : point_igrids,
			 "cell_indices" : point_icells,
			 "levels" : point_ilevels }


def cell_centers_from_grid_centers(
		numpy.ndarray[double, ndim=2] grid_centers,
		numpy.ndarray[int,    ndim=1] grid_levels):
	"""Compute the (ngrid, twotondim, ndim) array of the coordinates of the
	centers of the RAMSES cells from the grid centers and corresponding grid
	levels
	"""

	cdef int ngrids = grid_centers.shape[0]
	cdef int ndim = grid_centers.shape[1]
	cdef int twotondim = 1 << ndim

	# Allocate output array
	cdef numpy.ndarray[double, ndim=3] cell_centers = \
			numpy.empty([ngrids, twotondim, ndim])

	# Precompute the cell shifts
	cdef numpy.ndarray[double, ndim=2] xc = numpy.empty([twotondim, ndim])
	cdef int idim, icell, ind, igrid
	cdef int ishift
	for ind in range(twotondim):
		for idim in range(ndim):
			ishift = (ind >> idim) & 1
			xc[ind, idim] = ishift - 0.5

	# Compute cell centers
	cdef double dx
	for igrid in range(ngrids):
		for ind in range(twotondim):
			for idim in range(ndim):
				dx = 0.5**grid_levels[igrid]
				cell_centers[igrid, ind, idim] = grid_centers[igrid, idim] \
						+ dx * xc[ind, idim]

	return cell_centers



def grid_levels(amr_struct):
	levels = numpy.empty(amr_struct["ngrids"], 'i')

	# At each level, we store the grids of all CPUs, then all boundaries
	ngrids_level = amr_struct["ngridlevel"].sum(axis=0) + \
			amr_struct["ngridbound"].sum(axis=0)
	
	offset = 0
	for ilevel in range(len(ngrids_level)):
		ngrids = ngrids_level[ilevel]
		if ngrids > 0:
			levels[offset:(offset+ngrids)] = ilevel + 1
			offset += ngrids

	return levels



#@cython.boundscheck(False)
def amr2array_3d(amr_struct,
		numpy.ndarray[double, ndim=2] var_array, int icpu,
		xmin, xmax, int cubelevel, dest=None):
	""" Low-level AMR to cartesian array routine.

	Parameters:

		amr_struct -- a dictionary tuple containing the AMR structure, such as
			the output of read_ramses_amr_file()[1]

		ivar_array -- an (igrid, ind_cell) array of floats holding the cell data
			to map

		icpu -- the CPU to map (indexed from 0)

		xmin -- the lower coordinates of the cube

		xmax -- the upper coordinates of the box

		cubelevel -- the resolution level for the cube

		dest -- if specified, the array dest will be filled inplace
			instead of allocating a new array

	Returns: xmin_out, xmax_out, cube

		xmin_out -- adjusted (real) lower coordinates of the output cube

		xmax_out -- adjusted (real) upper coordinates of the output cube

		cube -- the output cartesian grid data
	"""

	cdef int ndim = amr_struct["ndim"]
	cdef int twotondim = 1 << ndim

	assert(ndim == 3)

	cdef double cube_dx = 0.5**cubelevel
	cdef numpy.ndarray[int, ndim=2] ngridlevel = amr_struct["ngridlevel"]
	cdef numpy.ndarray[int, ndim=2] ngridbound = amr_struct["ngridbound"]
	cdef int ngrids = amr_struct["ngrids"]
	cdef int ncpu = amr_struct["ncpu"]
	cdef int nboundary = amr_struct["nboundary"]

	cdef numpy.ndarray[double, ndim=3] xc = cell_centers_from_grid_centers(
			amr_struct["grid_centers"],
			grid_levels(amr_struct))

	cdef numpy.ndarray[int, ndim=2] son = amr_struct["son_indices"]

	cdef int ixmin, iymin, izmin
	cdef int ixmax, iymax, izmax
	cdef int ix, iy, iz

	cdef int ixmincube = int(xmin[0]/cube_dx)
	cdef int iymincube = int(xmin[1]/cube_dx)
	cdef int izmincube = int(xmin[2]/cube_dx)
	cdef int ixmaxcube = int(xmax[0]/cube_dx)
	cdef int iymaxcube = int(xmax[1]/cube_dx)
	cdef int izmaxcube = int(xmax[2]/cube_dx)

	# Allocate the cube
	cdef int nxcube = ixmaxcube-ixmincube
	cdef int nycube = iymaxcube-iymincube
	cdef int nzcube = izmaxcube-izmincube

	cdef numpy.ndarray[double, ndim=3] cube
	if dest is not None:
		cube = dest
	else:
		cube = numpy.zeros([nxcube, nycube, nzcube], 'd')

	cdef double cell_dx, hcell_dx, cell_val
	cdef int ixcube, iycube, izcube
	cdef int x_ok, y_ok, z_ok

	# Loop over grids
	cdef int nfilled = 0
	
	cdef int kcpu
	cdef int ind
	cdef int ilevel = 0
	cdef int igrid = 0
	cdef int ngridsremain = 0
	cell_dx = 0.5

	while(ilevel<cubelevel):

		# Skip all CPUs 0..icpu-1 for this level
		for kcpu in range(0, icpu):
			igrid += ngridlevel[kcpu, ilevel]

		# Loop over the CPU grids of this level
		ngridsremain = ngridlevel[icpu, ilevel]
		while(ngridsremain > 0):

			# Loop over cells
			for ind in range(twotondim):

				# Only stop at leaf cells
				if (son[igrid, ind] >= 0 and ilevel<cubelevel-1): continue

				hcell_dx = 0.5*cell_dx
				ixmin = <int>((xc[igrid, ind, 0] - hcell_dx)/cube_dx)
				ixmax = <int>((xc[igrid, ind, 0] + hcell_dx)/cube_dx)
				iymin = <int>((xc[igrid, ind, 1] - hcell_dx)/cube_dx)
				iymax = <int>((xc[igrid, ind, 1] + hcell_dx)/cube_dx)
				izmin = <int>((xc[igrid, ind, 2] - hcell_dx)/cube_dx)
				izmax = <int>((xc[igrid, ind, 2] + hcell_dx)/cube_dx)

				x_ok = (ixmin <= ixmaxcube) and (ixmax >= ixmincube)
				y_ok = (iymin <= iymaxcube) and (iymax >= iymincube)
				z_ok = (izmin <= izmaxcube) and (izmax >= izmincube)
				if not (x_ok and y_ok and z_ok):
					continue

				cell_val = var_array[igrid, ind]

				for ix in range(ixmin, ixmax):
					ixcube = ix - ixmincube
					if (ixcube<0 or ixcube>=nxcube): continue
					for iy in range(iymin, iymax):
						iycube = iy - iymincube
						if (iycube<0 or iycube>=nycube): continue
						for iz in range(izmin, izmax):
							izcube = iz - izmincube
							if (izcube<0 or izcube>=nzcube): continue
							cube[ixcube, iycube, izcube] = cell_val
							nfilled += 1

			# Go to next grid at this CPU and level
			ngridsremain -= 1
			igrid += 1

		# Skip all CPUs icpu+1, ... ncpu for this level
		for kcpu in range(icpu+1, ncpu):
			igrid += ngridlevel[kcpu, ilevel]

		# Skip all boundary grids for this level
		for kcpu in range(nboundary):
			igrid += ngridbound[kcpu, ilevel]

		ilevel += 1
		cell_dx *= 0.5

	return cube


