# License:
#   Copyright (C) 2011 Thomas GUILLET, Damien CHAPON, Marc LABADENS. All Rights Reserved.
#
#   This file is part of PyMSES.
#
#   PyMSES is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   PyMSES is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with PyMSES.  If not, see <http://www.gnu.org/licenses/>.
""" point_utils.pyx -- utility functions to generate arrays of points
"""

import numpy
cimport numpy
cimport cython
@cython.boundscheck(False)


def meshgrid(coords_by_axis):
	""" Constructs an array listing the points whose coordinates are formed by
	the cartesian product of the arrays listed in coords_by_axis.

	The points are listed in C (row-major) order.
	"""

	# Ensure coords_by_axis is a list of proper numpy arrays
	coords_by_axis = [
			numpy.asarray(ax, 'd') for ax in coords_by_axis ]

	cdef int ndim = len(coords_by_axis)

	if ndim == 1:
		return numpy.asarray(coords_by_axis[0])

	# Shape and stride arrays
	cdef numpy.ndarray[int, ndim=1] shape = numpy.array(
			[ax.size for ax in coords_by_axis], 'i')
	cdef long npoints = numpy.prod(shape)

	cdef numpy.ndarray[long, ndim=1] nperiods = numpy.cumprod(shape) / shape
	cdef numpy.ndarray[long, ndim=1] nrepeats = npoints/(nperiods * shape)

	# Output array
	cdef numpy.ndarray[double, ndim=2] points = numpy.zeros([npoints, ndim])

	cdef long ipoint, iperiod, irepeat
	cdef int idim, icoord

	cdef numpy.ndarray[double, ndim=1] coord_array

	# For every coordinate...
	for idim in range(ndim):

		# Get all the values which the coord idim takes
		coord_array = coords_by_axis[idim] # slow, but in outer loop

		ipoint = 0
		icoord = 0
		for iperiod in range(nperiods[idim]):
			for icoord in range(shape[idim]):
				for irepeat in range(nrepeats[idim]):
					points[ipoint, idim] = coord_array[icoord]
					ipoint += 1

	return points

__all__ = ["meshgrid"]
